@@ -3,12 +3,13 @@ import os
 import subprocess
 import sys
 from distutils.version import StrictVersion
+from itertools import chain
 
 import coverage
 import py
 import pytest
 import virtualenv
-from process_tests import TestProcess
+from process_tests import TestProcess as _TestProcess
 from process_tests import dump_on_error
 from process_tests import wait_for_strings
 
@@ -99,6 +100,29 @@ def test_foo(idx):
 pass
 '''
 
+SCRIPT_PARENT_CHANGE_CWD_IMPORT_CHILD = '''
+import subprocess
+import sys
+import os
+
+def pytest_generate_tests(metafunc):
+    for i in range(2):
+        metafunc.addcall(funcargs=dict(idx=i))
+
+def test_foo(idx):
+    os.mkdir("foobar")
+    os.chdir("foobar")
+
+    subprocess.check_call([
+        sys.executable,
+        '-c', 'import sys; sys.argv = ["", str(%s)]; import child_script' % idx
+    ])
+
+# there is a issue in coverage.py with multiline statements at
+# end of file: https://bitbucket.org/ned/coveragepy/issue/293
+pass
+'''
+
 SCRIPT_FUNCARG = '''
 import coverage
 
@@ -134,6 +158,8 @@ SCRIPT_RESULT = '9 * 89%'
 SCRIPT2_RESULT = '3 * 100%'
 CHILD_SCRIPT_RESULT = '[56] * 100%'
 PARENT_SCRIPT_RESULT = '8 * 100%'
+DEST_DIR = 'cov_dest'
+REPORT_NAME = 'cov.xml'
 
 xdist = pytest.mark.parametrize('opts', ['', '-n 1'], ids=['nodist', 'xdist'])
 
@@ -170,6 +196,96 @@ def test_annotate(testdir):
     assert result.ret == 0
 
 
+def test_annotate_output_dir(testdir):
+    script = testdir.makepyfile(SCRIPT)
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=annotate:' + DEST_DIR,
+                               script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        'Coverage annotated source written to dir ' + DEST_DIR,
+        '*10 passed*',
+    ])
+    dest_dir = testdir.tmpdir.join(DEST_DIR)
+    assert dest_dir.check(dir=True)
+    assert dest_dir.join(script.basename + ",cover").check()
+    assert result.ret == 0
+
+
+def test_html_output_dir(testdir):
+    script = testdir.makepyfile(SCRIPT)
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=html:' + DEST_DIR,
+                               script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        'Coverage HTML written to dir ' + DEST_DIR,
+        '*10 passed*',
+    ])
+    dest_dir = testdir.tmpdir.join(DEST_DIR)
+    assert dest_dir.check(dir=True)
+    assert dest_dir.join("index.html").check()
+    assert result.ret == 0
+
+
+def test_xml_output_dir(testdir):
+    script = testdir.makepyfile(SCRIPT)
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=xml:' + REPORT_NAME,
+                               script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        'Coverage XML written to file ' + REPORT_NAME,
+        '*10 passed*',
+    ])
+    assert testdir.tmpdir.join(REPORT_NAME).check()
+    assert result.ret == 0
+
+
+def test_term_output_dir(testdir):
+    script = testdir.makepyfile(SCRIPT)
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=term:' + DEST_DIR,
+                               script)
+
+    # backport of argparse to py26 doesn't display ArgumentTypeError message
+    result.stderr.fnmatch_lines([
+        '*argument --cov-report: *',
+    ] if tuple(sys.version_info[:2]) == (2, 6) else [
+        '*argument --cov-report: output specifier not supported for: "term:%s"*' % DEST_DIR,
+    ])
+    assert result.ret != 0
+
+
+def test_term_missing_output_dir(testdir):
+    script = testdir.makepyfile(SCRIPT)
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=term-missing:' + DEST_DIR,
+                               script)
+
+    # backport of argparse to py26 doesn't display ArgumentTypeError message
+    result.stderr.fnmatch_lines([
+        '*argument --cov-report: *',
+    ] if tuple(sys.version_info[:2]) == (2, 6) else [
+        '*argument --cov-report: output specifier not supported for: '
+        '"term-missing:%s"*' % DEST_DIR,
+    ])
+    assert result.ret != 0
+
+
 def test_cov_min_100(testdir):
     script = testdir.makepyfile(SCRIPT)
 
@@ -221,7 +337,7 @@ def test_central_nonspecific(testdir):
     ])
 
     # multi-module coverage report
-    assert any(line.startswith('TOTAL ') for line in result.stdout.lines[-4:])
+    assert any(line.startswith('TOTAL ') for line in result.stdout.lines)
 
     assert result.ret == 0
 
@@ -304,6 +420,37 @@ def test_no_cov_on_fail(testdir):
     result.stdout.fnmatch_lines(['*1 failed*'])
 
 
+def test_no_cov(testdir):
+    script = testdir.makepyfile(SCRIPT)
+
+    result = testdir.runpytest('-vvv',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=term-missing',
+                               '--no-cov',
+                               '-rw',
+                               script)
+    result.stdout.fnmatch_lines_random([
+        'WARNING: Coverage disabled via --no-cov switch!',
+        '*1 warnings*' if pytest.__version__.startswith('2.7') else 'WCOV-U1*Coverage disabled via --no-cov switch!',
+    ])
+
+
+def test_cov_and_failure_report_on_fail(testdir):
+    script = testdir.makepyfile(SCRIPT + SCRIPT_FAIL)
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-fail-under=100',
+                               script)
+
+    result.stdout.fnmatch_lines_random([
+        '*10 failed*',
+        '*coverage: platform*',
+        '*FAIL Required test coverage of 100% not reached*',
+        '*assert False*',
+    ])
+
+
 def test_dist_combine_racecondition(testdir):
     script = testdir.makepyfile("""
 import pytest
@@ -318,15 +465,17 @@ def test_foo(foo):
     result = testdir.runpytest('-v',
                                '--cov=%s' % script.dirpath(),
                                '--cov-report=term-missing',
-                               '-n', '5',
+                               '-n', '5', '-s',
                                script)
     result.stdout.fnmatch_lines([
         '*- coverage: platform *, python * -*',
         'test_dist_combine_racecondition* 2002 * 0 * 100%*',
         '*1000 passed*'
     ])
-    for line in result.stdout.lines:
+
+    for line in chain(result.stdout.lines, result.stderr.lines):
         assert 'The following slaves failed to return coverage data' not in line
+        assert 'INTERNALERROR' not in line
     assert result.ret == 0
 
 
@@ -414,6 +563,31 @@ parallel = true
     assert result.ret == 0
 
 
+def test_central_subprocess_change_cwd_with_pythonpath(testdir, monkeypatch):
+    stuff = testdir.mkdir('stuff')
+    parent_script = stuff.join('parent_script.py')
+    parent_script.write(SCRIPT_PARENT_CHANGE_CWD_IMPORT_CHILD)
+    stuff.join('child_script.py').write(SCRIPT_CHILD)
+    testdir.makefile('', coveragerc="""
+[run]
+branch = true
+parallel = true
+""")
+
+    monkeypatch.setitem(os.environ, 'PYTHONPATH', str(stuff))
+    result = testdir.runpytest('-vv', '-s',
+                               '--cov=child_script',
+                               '--cov-config=coveragerc',
+                               '--cov-report=term-missing',
+                               parent_script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        '*child_script* %s*' % CHILD_SCRIPT_RESULT,
+    ])
+    assert result.ret == 0
+
+
 def test_central_subprocess_no_subscript(testdir):
     script = testdir.makepyfile("""
 import subprocess, sys
@@ -487,7 +661,7 @@ def test_dist_subprocess_not_collocated(testdir, tmpdir):
     assert result.ret == 0
 
 
-def test_empty_report(testdir):
+def test_invalid_coverage_source(testdir):
     script = testdir.makepyfile(SCRIPT)
 
     result = testdir.runpytest('-v',
@@ -496,14 +670,27 @@ def test_empty_report(testdir):
                                script)
 
     result.stdout.fnmatch_lines([
-        '*- coverage: platform *, python * -*',
         '*10 passed*'
     ])
-    assert result.ret == 0
+    result.stderr.fnmatch_lines([
+        'Coverage.py warning: No data was collected.'
+    ])
+
+    if StrictVersion(coverage.__version__) <= StrictVersion("3.8"):
+        # older `coverage report` doesn't error on missing imports
+        assert result.ret == 0
+    else:
+        # newer `coverage report` errors on missing importts
+        result.stderr.fnmatch_lines([
+            'ERROR: Failed to generate report: No data to report.',
+        ])
+        assert result.ret != 0
+
     matching_lines = [line for line in result.outlines if '%' in line]
     assert not matching_lines
 
 
+@pytest.mark.skipif("'dev' in pytest.__version__")
 def test_dist_missing_data(testdir):
     venv_path = os.path.join(str(testdir.tmpdir), 'venv')
     virtualenv.create_environment(venv_path)
@@ -621,7 +808,7 @@ def test_cover_looponfail(testdir, monkeypatch):
     testdir.makeconftest(CONFTEST)
     script = testdir.makepyfile(BASIC_TEST)
 
-    monkeypatch.setattr(testdir, 'run', lambda *args: TestProcess(*map(str, args)))
+    monkeypatch.setattr(testdir, 'run', lambda *args: _TestProcess(*map(str, args)))
     with testdir.runpytest('-v',
                            '--cov=%s' % script.dirpath(),
                            '--looponfail',
@@ -698,6 +885,52 @@ def test_coveragerc_dist(testdir):
         ['test_coveragerc_dist* %s' % EXCLUDED_RESULT])
 
 
+SKIP_COVERED_COVERAGERC = '''
+[report]
+skip_covered = True
+
+'''
+
+SKIP_COVERED_TEST = '''
+
+def func():
+    return "full coverage"
+
+def test_basic():
+    assert func() == "full coverage"
+
+'''
+
+SKIP_COVERED_RESULT = '1 file skipped due to complete coverage.'
+
+
+@pytest.mark.skipif('StrictVersion(coverage.__version__) < StrictVersion("4.0")')
+@pytest.mark.parametrize('report_option', [
+    'term-missing:skip-covered',
+    'term:skip-covered'])
+def test_skip_covered_cli(testdir, report_option):
+    testdir.makefile('', coveragerc=SKIP_COVERED_COVERAGERC)
+    script = testdir.makepyfile(SKIP_COVERED_TEST)
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=%s' % report_option,
+                               script)
+    assert result.ret == 0
+    result.stdout.fnmatch_lines([SKIP_COVERED_RESULT])
+
+
+@pytest.mark.skipif('StrictVersion(coverage.__version__) < StrictVersion("4.0")')
+def test_skip_covered_coveragerc_config(testdir):
+    testdir.makefile('', coveragerc=SKIP_COVERED_COVERAGERC)
+    script = testdir.makepyfile(SKIP_COVERED_TEST)
+    result = testdir.runpytest('-v',
+                               '--cov-config=coveragerc',
+                               '--cov=%s' % script.dirpath(),
+                               script)
+    assert result.ret == 0
+    result.stdout.fnmatch_lines([SKIP_COVERED_RESULT])
+
+
 CLEAR_ENVIRON_TEST = '''
 
 import os
@@ -746,9 +979,26 @@ def test_dist_boxed(testdir):
     assert result.ret == 0
 
 
+@pytest.mark.skipif('sys.platform == "win32"')
+def test_dist_bare_cov(testdir):
+    script = testdir.makepyfile(SCRIPT_SIMPLE)
+
+    result = testdir.runpytest('-v',
+                               '--cov',
+                               '-n', '1',
+                               script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        'test_dist_bare_cov* %s*' % SCRIPT_SIMPLE_RESULT,
+        '*1 passed*'
+    ])
+    assert result.ret == 0
+
+
 def test_not_started_plugin_does_not_fail(testdir):
     plugin = pytest_cov.plugin.CovPlugin(None, None, start=False)
-    plugin.pytest_sessionfinish(None, None)
+    plugin.pytest_runtestloop(None)
     plugin.pytest_terminal_summary(None)
 
 
@@ -876,4 +1126,3 @@ def test_do_not_append_coverage(testdir, opts):
         'test_1* 0%',
         'test_2* %s*' % SCRIPT2_RESULT,
     ])
-
